import std/[json, options, os, paths, strutils]
import cliutils, docopt, illwill, mpv

import wdiwtlt/[cliconstants, commandbuffer, logging, medialibrary, models,
                scrolltext]

type
  CliMode {.pure.} = enum Direct, Command

  ScrollingDisplays* = ref object
    playing: ScrollText
    status: ScrollText

  History = ref object
    entries: seq[string]
    idx: int

  CliContext = ref object
    cfg: CombinedConfig
    cmd: CommandBuffer
    curMediaFile: Option[MediaFile]
    frame: int
    lib: MediaLibrary
    log: Option[Logger]
    mode: CliMode
    mpv: ptr handle
    stop*: bool
    tb: TerminalBuffer
    d: ScrollingDisplays

var ctx {.threadvar.}: CliContext

proc initMpv(): ptr handle =
  result = mpv.create()

  if result.isNil:
    raise newException(ValueError, "failed creating mpv context")

  # Enable default key bindings, so the user can actually interact with
  # the player (and e.g. close the window).
  result.set_option("terminal", "no")
  result.set_option("input-default-bindings", "yes")
  result.set_option("input-vo-keyboard", "yes")
  result.set_option("osc", false)
  #result.set_property("volume", "50.0")
  discard result.request_log_messages("no")

  # Done setting up options.
  check_error result.initialize()

proc cleanup() =
  terminate_destroy(ctx.mpv)
  illWillDeinit()
  showCursor()
  echo ""

proc exitHandlerHook() {.noconv.} =
  # This is called when the user presses Ctrl+C
  # or when the program exits normally.
  echo "Ctrl-C"
  terminate_destroy(ctx.mpv)
  illWillDeinit()
  quit(QuitSuccess)

proc setMode(ctx: var CliContext, mode: CliMode) =
  ctx.mode = mode
  case mode
  of CliMode.Direct:
    hideCursor()
    ctx.tb.clear()
  of CliMode.Command:
    showCursor()
    ctx.cmd.mode = EditMode.Insert

proc initContext(args: Table[string, Value]): CliContext =
  var wdiwtltCfgFilename: string

  let log = getLogger("wdiwtlt")

  let cfgLocations =
    if args["--config"]: @[$args["--config"]]
    elif existsEnv("XDG_CONFIG_HOME"): @[getEnv("XDG_CONFIG_HOME") / "wdiwtlt.json"]
    else: @[$getEnv("HOME") / ".config" / "wdiwtlt.json" ]

  try: wdiwtltCfgFilename = findConfigFile(".wdiwtlt.json", cfgLocations)
  except ValueError:
    log.error "could not find wdiwtlt config file: " & wdiwtltCfgFilename
    if not args["--config"]:
      wdiwtltCfgFilename = cfgLocations[0]

      var cfgFile: File
      try:
        cfgFile = open(wdiwtltCfgFilename, fmWrite)
        cfgFile.write((%*{
          "libraryRoot":
            if args["--library-root"]:
              $args["--library-root"]
            else:
              getEnv("XDG_MUSIC_DIR", getEnv("HOME") / "music"),
          "dbPath": getEnv("XDG_DATA_HOME", (getEnv("HOME") / ".local/share")) / "wdiwtlt/db.sqlite",
        }).pretty)
        log.info "created sample config file at " & wdiwtltCfgFilename
      except CatchableError:
        log.error "could not write default .wdiwtlt.json to " &
          wdiwtltCfgFilename
      finally: close(cfgFile)

    quit(QuitFailure)

  log.debug("loading config from '$#'" % [wdiwtltCfgFilename])
  let cfg = initCombinedConfig(wdiwtltCfgFilename, args)

  result = CliContext(
    cfg: cfg,
    cmd: initCommandBuffer(),
    curMediaFile: none(MediaFile),
    d: ScrollingDisplays(
      playing: initScrollText("Nothing playing", terminalWidth() - 1),
      status: initScrollText("Idle", terminalWidth() - 1)),
    frame: 0,
    lib: initMediaLibrary(
      rootDir = getVal(cfg, "libraryRoot").Path,
      dbPath = getVal(cfg, "dbPath").Path),
    log: log,
    mode: CliMode.Direct,
    mpv: initMpv(),
    stop: false,
    tb: newTerminalBuffer(terminalWidth(), terminalHeight()))

  if logService.isSome:
    let customLogAppender = initCustomLogAppender(doLogMessage =
      proc (msg: LogMessage): void =
        ctx.tb.write(0, 0, msg.message))

    logService.get.clearAppenders()
    logService.get.addAppender(customLogAppender)

proc render(ctx: CliContext) =
  if ctx.frame mod 20 == 0: discard ctx.d.playing.nextTick
  ctx.tb.write(0, 0, ctx.d.playing)

  if ctx.mode == CliMode.Command:
    case ctx.cmd.mode:
    of EditMode.Insert:     stdout.write("\x1b[5 q")
    of EditMode.Overwrite:  stdout.write("\x1b[0 q")
    else:
      stdout.write("\x1b[2 q")
    ctx.tb.write(0, 1, ":$#" % [$ctx.cmd])

  ctx.tb.display()

proc mainLoop(ctx: var CliContext) =

  hideCursor()

  while not ctx.stop:
    let key = getKey()

    if ctx.mode == CliMode.Direct:
      case key
      of Key.Q:     ctx.stop = true
      of Key.Colon, Key.I: ctx.setMode(CliMode.Command)
      else: discard

    elif ctx.mode == CliMode.Command:
      ctx.tb.write(0, 1, ":", ' '.repeat(($ctx.cmd).len))
      case key
      of Key.Enter:
        let command = $ctx.cmd
        ctx.cmd.clear
        # TODO: process command
      of Key.Backspace: ctx.cmd.handleInput(Key.Backspace)
      of Key.Escape:
        if ctx.cmd.mode == EditMode.Command:
          ctx.setMode(CliMode.Direct)
          ctx.cmd.clear
        else: ctx.cmd.handleInput(Key.Escape)
      else: ctx.cmd.handleInput(key)

    render(ctx)

    # target 50 FPS
    sleep(20)
    ctx.frame = (ctx.frame + 1 mod 50)

when isMainModule:

  discard enableLogging()
  try:
    let args = docopt(USAGE, version=VERSION)

    illWillInit(fullscreen=true)
    ctx = initContext(args)

    mainLoop(ctx)

  except Exception:
    let ex = getCurrentException()
    getLogger("wdiwtlt").error(%*{
      "msg": "Unhandled exception",
      "error": ex.msg,
      "trace": ex.getStackTrace()
    })
    cleanup()
    quit(QuitFailure)
  finally:
    cleanup()
